import pygame
import os

def union(settuple):
    return set().union(*settuple)

def imgpath(filename):
    return os.path.join('img', filename)

def levpath(filename):
    return os.path.join('levels', filename)

def audpath(filename):
    return os.path.join('sound', filename)

NROWS = 16
NCOLS = 24
TILESIZE = 40
PCIMGHEIGHT = 53
ROBOTSPEED = 2
PCSPEED = 2
SLOWSPEED = 0.4
SLUDGESPEED = 0.2
SLUDGEMAXRADIUS = 200
WINSIZE = (TILESIZE * NCOLS, TILESIZE * NROWS)
WINDOW = pygame.display.set_mode(WINSIZE, 0, 32)
pygame.display.set_caption('Mischief Moon')
EFFECTSURF = pygame.Surface(WINSIZE, pygame.SRCALPHA)
WHITEOUTSURF = pygame.Surface(WINSIZE)
OCCUPYSIZE = TILESIZE - 4

# Seconds between frames for animations that run while engine is stalled
ANIMATIONINTERVAL = 0.25
FIREANIM = [pygame.image.load(imgpath('fire.png')),
            pygame.image.load(imgpath('fire_2.png'))]
GOALANIM = [pygame.image.load(imgpath('stars.png')),
            pygame.image.load(imgpath('stars_2.png'))]

SPACEMANBIG = pygame.image.load(imgpath('spacemanbig.png'))
BACKGROUND = pygame.image.load(imgpath('moonsurf.png'))
TILENAMES = {'f':'floor.png',
            '_':'floor.png',
            '/':'floornobot.png',
            'w':'wall.png',
#            'x':'hazard.png',
#            'g':'goal.png',
            '=':'pipehoriz.png',
            '|':'pipevert.png',
            '+':'bursthoriz.png',
            '\\':'burstvert.png',
#            '.':'',
#            '1':'lever.png',
#            '2':'lever.png',
#            '3':'lever.png',
#            '4':'lever.png',
#            '5':'lever.png',
            '6':'closeddoor.png',
            '7':'closeddoor.png',
            '8':'closeddoor.png',
            '9':'closeddoor.png',
            '0':'closeddoor.png',
            '^':'opendoor.png',
            '&':'opendoor.png',
            '*':'opendoor.png',
            '(':'opendoor.png',
            ')':'opendoor.png'
             }
GOALIMG = 'goal.png'
SWITCHIMG = pygame.image.load(imgpath('switch.png'))
BROKENSWITCHIMG = pygame.image.load(imgpath('switchbroken.png'))


OUTFILENAME = 'progress.dat'
STARTLEVEL = 1
BOMBAVAILABLE = 21

MAXLEVEL = 29

EMPTY = '_'

SCORCHED = 'x'      # Tile to use for scorched earth after bomb goes off

PIPES = ('=', '|')
BURSTPIPES = ('+','\\')
PIPE2BURST = {'=':'+', '|':'\\'}
BURST2PIPE = {'+':'=', '\\':'|'}

# Levers should be the numbers 1-5.
# Doors controlled by those levers are 6-0. If they show up as their SHIFTed symbol, then that door is intially open.
LEVERS = ('1','2','3','4','5')
BREAKSWITCHES = ('!','@','#','$','%')
OPENDOORS = ('^','&','*','(',')')
CLOSEDDOORS = ('6','7','8','9','0')
LEVERDOOR = {'1':'6', '2':'7', '3':'8', '4':'9', '5':'0',
            '!':'6', '@':'7', '#':'8', '$':'9', '%':'0'}
LEVEROPENDOOR = {'1':'^', '2':'&', '3':'*', '4':'(', '5':')',
            '!':'^', '@':'&', '#':'*', '$':'(', '%':')'}

GOAL = 'g'

BLOCKED = union((set(['w','.']), PIPES, CLOSEDDOORS))
BOTBLOCKED = union((BLOCKED,set('/') ))
CLEAR = union((set(['_','f','/']), OPENDOORS, LEVERS, GOAL, BURSTPIPES))
ALLTILES = BLOCKED.union(CLEAR)
#BLOCKED = {'w','=','|','.'}.union(CLOSEDDOORS)
#CLEAR = {'_','f'}.union(OPENDOORS).union(LEVERS).union(GOAL)

BOMBTIMER = 240

LEVELMESSAGES = {3: 'Press B to drop a bomb. Bombs will explode unless the robot disarms them.'}
MESSAGELOC = (TILESIZE, (NROWS-1) * TILESIZE)
LEVELTAGLOC = (WINSIZE[0] - 100, 10)
MYFONT = 'freesansbold.ttf'
FONTSIZE = 20

cTEXT = (255,255,255)
cINTROTEXT = (0,40,0)
cOUTLINE = (0,0,60)
cSLUDGE = (0, 100, 0, 128)

# For animation
PCFRAMESPERFRAME = 6

SOUNDNAMES = ['burstpipe',
            'robotmove',
            'explode',
            'death',
            'lever',
            'switch',
            'door',
            'setbomb',
            'goal',
            'outofbombs',
            'disarmbomb',
            'repair']

# This was a nice comprehension but it broke under Python 2.6, so...
#SOUND = {name: pygame.mixer.Sound(audpath(name+'.ogg')) for name in SOUNDNAMES}
SOUND = {'burstpipe':pygame.mixer.Sound(audpath('burstpipe'+'.ogg')),
            'robotmove':pygame.mixer.Sound(audpath('robotmove'+'.ogg')),
            'explode':pygame.mixer.Sound(audpath('explode'+'.ogg')),
            'death':pygame.mixer.Sound(audpath('death'+'.ogg')),
            'lever':pygame.mixer.Sound(audpath('lever'+'.ogg')),
            'switch':pygame.mixer.Sound(audpath('switch'+'.ogg')),
            'door':pygame.mixer.Sound(audpath('door'+'.ogg')),
            'setbomb':pygame.mixer.Sound(audpath('setbomb'+'.ogg')),
            'goal':pygame.mixer.Sound(audpath('goal'+'.ogg')),
            'outofbombs':pygame.mixer.Sound(audpath('outofbombs'+'.ogg')),
            'disarmbomb':pygame.mixer.Sound(audpath('disarmbomb'+'.ogg')),
            'repair':pygame.mixer.Sound(audpath('repair'+'.ogg'))}
