from functions import *
import pygame, string, time
from pygame.locals import *

# Things handled by the game engine (movement, victory, starting levels, etc.)
class Game():

    curlevel = STARTLEVEL 
    customlevs = False
    won = False

    def __init__(self, startlev = False):
        if startlev:
            if startlev == 'c':
                self.customlevs = True
                self.curlevel = 1
            else:
                self.curlevel = startlev

        self.goal = Goal((0,0))
        self.hazards = []
        self.levers = []
        self.switches = []
        self.effects = []
        self.bombused = False
        self.text = pygame.font.Font(MYFONT, FONTSIZE)
        self.outlinetext = pygame.font.Font(MYFONT, FONTSIZE)
        self.read_level()
        EFFECTSURF.fill((255,255,255,0))

    def step(self):
        self.read_inputs()
        self.draw_all()
        pygame.display.update()

        for bot in self.bots:
            bot.behave(self)
        for effect in self.effects:
            effect.update()
        for switch in self.switches:
            if self.is_dead(switch) and switch.intact:
                switch.destroy(self)
                self.fix_matching_levers(switch)

        # Bomb class will handle explosion
        self.bomb.tick(self)

        self.death_conditions()
        self.victory_conditions()
        return self.won

    def read_inputs(self):
        allpressed = pygame.key.get_pressed()
        inputs =   {K_b: self.drop_bomb,
                    K_r: self.reset_level,
                    K_SPACE: self.use_object,
                    K_v: self.smash_around}
        for key in inputs:
            if allpressed[key]:
                inputs[key]()

        L = allpressed[K_LEFT]
        R = allpressed[K_RIGHT]
        U = allpressed[K_UP]
        D = allpressed[K_DOWN]

        if L and not R:
            if U and not D:
                self.move_char('l', reduced = True)
                self.move_char('u', reduced = True)
            elif D:
                self.move_char('l', reduced = True)
                self.move_char('d', reduced = True)
            else:
                self.move_char('l')
        elif R:
            if U and not D:
                self.move_char('r', reduced = True)
                self.move_char('u', reduced = True)
            elif D:
                self.move_char('r', reduced = True)
                self.move_char('d', reduced = True)
            else:
                self.move_char('r')
        elif U and not D:
            self.move_char('u')
        elif D:
            self.move_char('d')
        
    def would_put_in_occupied(self, d, startxy, speed):
        """Would this move put me in an occupied tile?"""
        speed = int(speed+1)
        endtile = functions.xy2tile(functions.d2xy(d, startxy,speed))
        if endtile in BLOCKED or self.tile_is_occupied(endtile):
            return True
        return False

    def draw_all(self):
        self.board.draw('clear')
        self.draw_things()
        self.draw_effects()
        self.draw_hazards()
        WINDOW.blit(EFFECTSURF, (0,0))
        self.board.draw('blocked')
        self.goal.draw()
        self.draw_characters()
        self.show_level_message()
        self.show_level_tag()

    def read_level(self):

        # General setup operations

        self.board = Board()
        self.pc = PlayerCharacter()
        self.bots = []
        self.bomb = Bomb()

        # Read from level file
    
    
        if self.customlevs:
            try:
                f = open(levpath('custom' + str(self.curlevel) + '.txt'))
            except:
                print 'All out of custom levels!'
                self.won = True
        else:
            try:
                f = open(levpath('level' + str(self.curlevel) + '.txt'))
            except:
                try:
                    f = open(levpath('level1.txt'))
                    try:
                        os.remove(os.path.join(OUTFILENAME))
                    except:
                        pass
                except:
                    print 'Failed to open level', self.curlevel
                    self.quit()
#        tileletters = [[line.strip()[x] for x in xrange(len(line.strip()))] for line in f]
        try:
            wholefile = f.read().split('\n')
        except:
            return
        tileletters = [[wholefile[line].strip()[x] for x in range(NCOLS)] for line in range(NROWS)]
        try:
            self.msg = wholefile[NROWS]
        except:
            self.msg = ''
        f.close()

        levers = []
        doors = []

        for row in xrange(len(tileletters)):
            for col in xrange(len(tileletters[row])):
                l = tileletters[row][col]
                if l in string.ascii_uppercase:
                    self.init_character_pos(l, (row,col))
                    tileletters[row][col] = EMPTY
                if l == SCORCHED:
                    self.hazards.append(Scorched((row,col)))
                    tileletters[row][col] = EMPTY
                if l == GOAL:
                    self.goal = Goal((row,col))
                    self.goal.img = pygame.image.load(imgpath(GOALIMG))
                    tileletters[row][col] = EMPTY
                if l in LEVERS or l in BREAKSWITCHES:
                    levers.append(((row,col),l))
                    tileletters[row][col] = EMPTY
                if l in LEVERDOOR.values() or l in LEVEROPENDOOR.values():
                    doors.append(((row,col),l))
        self.board.board = tileletters

        for lever in levers:
            if lever[1] in LEVERS:
                self.levers.append(Lever(lever[0], lever[1]))
            elif lever[1] in BREAKSWITCHES:
                self.switches.append(DestructibleSwitch(lever[0], lever[1]))
        
        for switchgrp in (self.switches, self.levers):
            for n in xrange(len(switchgrp)):
                for door in doors:
                    if door[1] == LEVERDOOR[switchgrp[n].value]:
                        switchgrp[n].add_door(door[0])
                    elif door[1] == LEVEROPENDOOR[switchgrp[n].value]:
                        switchgrp[n].add_door(door[0], True)

    def show_level_message(self):
        textsurf = self.text.render(self.msg, True, cTEXT)
        outlinesurf = self.outlinetext.render(self.msg, True, cOUTLINE)
        WINDOW.blit(outlinesurf, (MESSAGELOC[0]+2, MESSAGELOC[1]+2))
        WINDOW.blit(outlinesurf, (MESSAGELOC[0]-1, MESSAGELOC[1]-1))
#        WINDOW.blit(outlinesurf, (MESSAGELOC[0]+1, MESSAGELOC[1]-1))
#        WINDOW.blit(outlinesurf, (MESSAGELOC[0]-1, MESSAGELOC[1]+1))
        WINDOW.blit(textsurf, MESSAGELOC)

    def show_level_tag(self):
        textsurf = self.text.render('Level ' + str(self.curlevel), True, cTEXT)
        WINDOW.blit(textsurf, LEVELTAGLOC)

    def draw_effects(self):
        for effect in self.effects:
            effect.draw()

    def draw_hazards(self):
        for h in self.hazards:
            h.draw()

    def remove_hazard(self, loc):
        for h in self.hazards:
            if h.get_tile() == loc:
                self.hazards.remove(h)

    def draw_characters(self):
        for ch in self.bots + [self.pc]:
            ch.draw()

    def draw_things(self):
        self.bomb.draw()
        for lever in self.levers + self.switches:
            lever.draw()

    def init_character_pos(self, letter, coord):
        if letter == 'P':
            self.pc.set_pos(coord)
        elif letter == 'R':
            self.bots.append(Robot(coord))
        elif letter == 'A':
            self.bots.append(RobotArm(coord))

    def place(self, graphic, coords):
        WINDOW.blit(graphic, coords)

    def place_tile(self, graphic, coords):
        self.place(graphic, (coords[1] * TILESIZE, coords[0] * TILESIZE))
        
    def move_char(self, d, reduced = False):
        # mod determines where the char WOULD end up if allowed to move
        mod = self.pc.speed
        if d == 'u' or d == 'l':
            mod *= -1
        
        if d == 'u' or d == 'd':
            yside = self.pc.get_side(d)
            xsides = (self.pc.get_side('l'), self.pc.get_side('r'))
            for xside in xsides:
                if (yside+mod)//TILESIZE not in range(NROWS):
                    return False
                totile = xy2tile((xside,yside + mod))
#                if self.board.tile_at_xy((xside,yside + mod)) in BLOCKED or self.tile_is_occupied(totile):
                if self.board[totile] in BLOCKED or self.tile_is_occupied(totile):
                    try:
                        passside = self.board.get_passable_side(totile, d, self.pc)
#                    self.move_char(passside, reduced = True)
                        passtile = functions.d2tile(passside,self.pc.get_tile())
#                        if not self.would_put_in_occupied(passside, self.pc.get_xy(), SLOWSPEED) and not self.board[passtile] in BLOCKED:
#                        if not self.would_put_in_occupied(passside, self.pc.get_xy(), SLOWSPEED):
                        if not self.tile_is_occupied(passtile, self.pc.get_tile()) and not self.board[passtile] in BLOCKED:
                            self.pc.move(passside, wraparound = True)
                    except:
                        pass
                    return False

        elif d == 'l' or d == 'r':
            xside = self.pc.get_side(d)
            ysides = (self.pc.get_side('u'), self.pc.get_side('d'))
            for yside in ysides:
                if (xside+mod)//TILESIZE not in range(NCOLS):
                    return False
                totile = xy2tile((xside + mod,yside))
                if self.board.tile_at_xy((xside + mod,yside)) in BLOCKED or self.tile_is_occupied(totile):
                    try:
                        passside = self.board.get_passable_side(totile, d, self.pc)
                        passtile = functions.d2tile(passside,self.pc.get_tile())
#                        if not self.would_put_in_occupied(passside, self.pc.get_xy(), SLOWSPEED) and not self.board[passtile] in BLOCKED:
                        if not self.tile_is_occupied(passtile, self.pc.get_tile()) and not self.board[passtile] in BLOCKED:
#                        if not self.would_put_in_occupied(passside, self.pc.get_xy(), SLOWSPEED):
                            self.pc.move(passside, wraparound = True)
                    except:
                        pass
                    return False
        self.pc.move(d, reduced)

    def use_object(self):
        for lever in self.levers:
            if lever.get_tile() == self.pc.get_tile():
                lever.pull(self)
                self.fix_matching_levers(lever)

    def fix_matching_levers(self, lever):
        # Set the door open value for other levers of this ID to match
        # Kludgy, but it's late
        for matchlever in self.levers + self.switches:
            if matchlever.value == lever.value:
                for i in range(len(matchlever.doors)):
                    matchlever.doors[i][1] = lever.doors[i][1]

    def smash_around(self):
        pctile = self.pc.get_tile()
        for pipetype in PIPES:
            adjpipe = self.board.adjacent_terrain(pctile, pipetype)
            if adjpipe:
                self.burst_pipe(adjpipe)
                break

    def burst_pipe(self, pipeloc):
        SOUND['burstpipe'].play()
        time.sleep(0.5)
        self.hazards.append(SpecialTile(pipeloc))
        self.board.change_tile(pipeloc, PIPE2BURST[self.board[pipeloc]])
        self.effects.append(Sludge(tile2xy(pipeloc)))

    def fix_pipe(self, pipeloc):
        for effect in self.effects:
            if effect.get_tile() == pipeloc and effect.is_active():
                effect.deactivate()
                self.board.change_tile(pipeloc, BURST2PIPE[self.board[pipeloc]])
        self.remove_hazard(pipeloc)
        time.sleep(0.5)

    def drop_bomb(self):
        if not self.bomb.live and (self.customlevs or self.bomb.is_available(self.curlevel)):
            if not self.bombused:
                self.bomb.arm(self.pc.get_tile())
                self.bombused = True

    def tile_is_occupied(self, tile, bybot = True):
        if not bybot:
            if self.pc.get_tile() == tile:
                return True
            return False
        for bot in self.bots:
            if bot.get_tile() == tile:
                return True
        return False


    def is_dead(self, what = False):
        if not what:
            what = self.pc

        for hazard in self.hazards:
            if what.get_tile() == hazard.get_tile():
                return True
        for effect in self.effects:
            if effect.explosion_contains(what.get_xy()):
                return True
        return False

    def has_won(self):
        if self.pc.get_tile() == self.goal.get_tile():
            return True
        return False

    def death_conditions(self):
        if self.is_dead():
            SOUND['death'].play()
            self.reset_level()

    def victory_conditions(self):
        if self.has_won():
            if self.curlevel >= MAXLEVEL:
                self.won = True
                return
            else:
                self.pass_level()
                return

    def pass_level(self):
        SOUND['goal'].play()
        time.sleep(1.5)
        self.curlevel += 1
        w = open(os.path.join(OUTFILENAME),'w')
        w.write(str(self.curlevel))
        w.close()
        
        self.__init__()

    def reset_level(self):
        time.sleep(1)
        self.__init__()

    def quit(self):
        pygame.quit()
        sys.exit()


class Board():

    def __init__(self):
        self.nukedspaces = []
        self.board = None
        self.tilemap = None
        # Surfaces for the CLEAR and BLOCKED spaces.
        # Draw to these only once to save time UNLESS something changes,
        #   in which case they'll need to be redrawn.
        self.clear = pygame.Surface(WINSIZE, pygame.SRCALPHA)
        self.blocked = pygame.Surface(WINSIZE, pygame.SRCALPHA)
        self.cleardrawn = False
        self.blockeddrawn = False

    def __getitem__(self, tup):
        return self.board[tup[0]][tup[1]]

    def draw(self, drawset = 'clear'):
        """ drawset will be 'clear' or 'blocked' coming from Game class """
        if drawset == 'clear':
            if not self.cleardrawn:
                WINDOW.blit(BACKGROUND, (0,0))
                self.redraw(self.clear, CLEAR)
                self.cleardrawn = True
            WINDOW.blit(self.clear, (0,0))
        elif drawset == 'blocked':
            if not self.blockeddrawn:
                self.redraw(self.blocked, BLOCKED)
                self.blockeddrawn = True
            WINDOW.blit(self.blocked, (0,0))
        
    def redraw(self, surf, drawset):
        if self.tilemap == None:
            self.load_tilemap_images()
        surf.fill((0,0,0,0))
        for row in xrange(NROWS):
            for col in xrange(NCOLS):
                if self.board[row][col] in drawset:
                    self.place(surf, self.tilemap[row][col], (col * TILESIZE, row * TILESIZE))

    def adjacent_terrain(self, (r,c), t):
        up = (r-1,c)
        down = (r+1,c)
        left = (r,c-1)
        right = (r,c+1)
        for d in (up, down, left, right):
            if self.board[d[0]][d[1]] == t:
                return d

    def get_passable_side(self, (r,c), dimension, pc):
        """Dimension is either 'l'/'r' or 'u'/'d'
        """
#        pctile = pc.get_tile()
        charxy = pc.get_xy()
        if dimension in ('l','r'):
            up = (r-1,c)
            down = (r+1,c)
            upc = functions.tile_center(up)
            downc = functions.tile_center(down)
            if self[up] in CLEAR and self[down] in CLEAR:
                if abs(charxy[1] - upc[1]) < abs(charxy[1] - downc[1]):
                    return 'u'
                else:
                    return 'd'
            elif self[up] in CLEAR:
                return 'u'
            elif self[down] in CLEAR:
                return 'd'
        elif dimension in ('u','d'):
            left = (r,c-1)
            right = (r,c+1)
            leftc = functions.tile_center(left)
            rightc = functions.tile_center(right)
            if self[left] in CLEAR and self[right] in CLEAR:
                if abs(charxy[0] - leftc[0]) < abs(charxy[0] - rightc[0]):
                    return 'l'
                else:
                    return 'r'
            elif self[left] in CLEAR:
                return 'l'
            elif self[right] in CLEAR:
                return 'r'
            

    def change_tile(self, (r,c), new):
        self.board[r][c] = new
        try:
            self.tilemap[r][c] = pygame.image.load(imgpath(TILENAMES[new]))
        except:
            self.tilemap[r][c] = pygame.Surface((0,0))
        self.cleardrawn = False
        self.blockeddrawn = False

    def load_tilemap_images(self):
        self.tilemap = [[a for a in b] for b in self.board]
        for row in xrange(len(self.tilemap)):
            for col in xrange(len(self.tilemap[row])):
                thistile = self.board[row][col]
                try:
                    self.tilemap[row][col] = pygame.image.load(imgpath(TILENAMES[thistile]))
                except:
                    self.tilemap[row][col] = pygame.Surface((0,0))

    def place(self, surf, graphic, coords):
        """Draw a single tile."""
        surf.blit(graphic, coords)

    def tile_at_xy(self, xy):
        row, col = (xy[1]//TILESIZE, xy[0]//TILESIZE)
        return self.board[row][col]

    def get_eligible_tiles(self, (r,c), agent = 'bot'):
        """ See which tiles are eligible to travel to.
            Possible values for agent could be 'bot', 'pc', or 'explosion'. """

        if agent == 'bot':
            blocked = BOTBLOCKED
        else:
            blocked = BLOCKED

        eligible = set((r, c) for r in xrange(len(self.board)) for c in xrange(len(self.board[r])) if self.board[r][c] not in blocked)
        return eligible
